' OpenSaveDialog2.bas



' *** Command line parameters:

' 1 "New Project" "Graphic files (*.png;*.bmp;*.jpg)|All files (*.*)|*.*" "c:\users" "picture.png" 1

' mode "title" "string for filters description and filters" "start directory" "filename" Show_hidden_flag

'   title (mandatory or "")

'   filters: (mandatory)
'       string to show selectable file filters (file extension)
'       Syntax:
'           "file type description | *.extension" (for single extension filetypes)
'           "files type description | *.ext1;*.ext2;*.ext3" (for multiple extension filetypes)

'   start directory: (optional)
'       start directory/folder to show in dialog box.

'   filename: (optional)
'       default filename for save and save as

'   Show_hidden_flag: (optional parameter)
'       Flag to show/no show hidden and system files and directories (True/False)
'       == 0 -> False      == 1 -> True

' All strings must be in quotes.

' mode:

'    - 6 options for dialog box type (title):
'        1="New"     (Title parameter is ignored. Option is "")
'        2="Open"    (Title parameter is ignored. Option is "")
'        3="Save"    (Title parameter is ignored. Option is "")
'        4="Save as" (Title parameter is ignored. Option is "")
'        5="Custom title for NEW/SAVE/SAVE AS"   (Title parameter is mandatory)
'        6="Custom title for OPEN"   (Title parameter is mandatory)

' Filename, start directory and show hidden flag are optional and can be parsed in any order.

' Feb 24 2022 - Public version released


' A universal open/save/save as configurable dialog box

' Input: It will accept config parameters similar to MS Windows Open/Save Dialog API functions
' Output: path+Filename+ext / (or empty string if cancelled or no file selected)

' It is able to:
'
'   Show folders and files in current path.
'   Alow to move in file list and filder list
'   Alow to click on any file name or folder name
'   If file name is clicked, it is shown in the File textbox
'   If folder name is double-clicked, then the path is changed and the textlists are updated
'   If a file name is double-clicked, then filename is selected and Accept button is clicked
'   If Accept button is clicked, then selected filename (or empty string if no filename selected) is returned
'   If Cancel button is clicked, then an empty string is returned.
'   If in mode Save or Save as, and an existing filename is selected, then it warns about overwriting.



': This program uses
': InForm - GUI library for QB64 - v1.2
': Fellippe Heitor, 2016-2020 - fellippe@qb64.org - @fellippeheitor
': https://github.com/FellippeHeitor/InForm
'-----------------------------------------------------------

': Controls' IDs: ------------------------------------------------------------------
Rem NOTICE: THIS FORM HAS BEEN RECENTLY EDITED
'>> The controls in the list below may have been added or renamed,
'>> and previously existing controls may have been deleted since
'>> this program's structure was first generated.
'>> Make sure to check your code in the events SUBs so that
'>> you can take your recent edits into consideration.
': ---------------------------------------------------------------------------------

Dim Shared OpenSaveForm As Long ' Form
Dim Shared PathTB As Long ' Textbox for current path
Dim Shared FolderListBox As Long ' listbox for folder list
Dim Shared FileListBox As Long ' listbox for file list
Dim Shared FileNameTextBox As Long ' textbox
Dim Shared FilterList As Long
Dim Shared DriveList As Long ' for drive change
Dim Shared AcceptButton As Long
Dim Shared CancelButton As Long


' ***
' General use variables
' ***
Dim Shared ShowHidden%% ' flag to show/no show hidden files and folders

Dim Shared ClickTimer As Double ' preparation for double click
Dim Shared ClickedControl As Long ' clicked control ID

Dim Shared DialogMode As _Unsigned _Byte ' mode for dialog box
Dim Shared StartDirectory$ ' start directory to show in dialog

Dim Shared FileName As String
Dim Shared path$
Dim Shared filter$(256)
ReDim Shared FileList$(256)
Dim Shared FilesCount As _Unsigned _Integer64 ' n files in a directory
Dim Shared DrivesCount As _Unsigned _Byte ' n drives (Windows only)
Dim Shared FilterList_Value As _Unsigned _Byte ' keeps number of last filter selected in dropdownlist box
Dim Shared FileList_Value As _Unsigned _Integer64 ' (8 bytes) keeps number of last filename selected in list

Const TempFile$ = "__OpenSaveDialog__.tmp" ' temporary file to store filenames/foldernames and chosen filename
Dim Shared TempDirFile$
Dim Shared TempPath$
Dim Shared ErrorCode As _Unsigned _Byte ' error flag to use when returning from error trap


$If WIN Then
    Const slash$ = "\"
$Else
    Const slash$ = "/"
$End If

' Variables for drive selection

$If WIN Then
    Dim Shared DriveList_Value As _Unsigned _Byte ' current selected item in drive list
    Dim Shared StartDrive As String ' stores drive where the program was started from
    Dim Shared StartDriveValue As _Unsigned _Byte ' start drive item in drive list
$End If


' ***
' Library to call Windows API
' ***

'Uses Kernel32 API to lists all available drives on system
' Inspired on:
' http://www.qb64.org/wiki/Windows_Libraries

$If WIN Then
    Declare Library
        Function GetLogicalDriveStringsA (ByVal nBuff As Long, lpbuff As String)
    End Declare


    ' ***
    ' Error trap for CHDIR
    ' ***

    DriveError:

    Select Case Err
        Case 0 ' to prevent error message at program start
            Exit Select

        Case 76
            DriveErrorMessage ' cannot use getitem$( function before  $ include
            Resume ' re-tries chdir

        Case Else
            UnhandledErrorMessage
            End
    End Select
$End If


PathError:
Select Case Err
    Case 0 ' to prevent error message at program start
        Exit Select

    Case 76
        PathErrorMessage ' cannot use Text(PathTB)=TempPath$ before  $ include ???
        ErrorCode = Err ' makes the BeforeInit sub to exit if not valid path parsed in command line
        StartDirectory$ = _StartDir$ ' prevents initial chdir to fail again
        Resume ' re-tries chdir

    Case Else
        UnhandledErrorMessage
        End
End Select



': External modules: ---------------------------------------------------------------
'$INCLUDE:'InForm\InForm.bi'
'$INCLUDE:'InForm\xp.uitheme'
'$INCLUDE:'OpenSaveDialog2.frm'



': Event procedures: ---------------------------------------------------------------


' ***
' This happens before drawing form window and initialize controls and variables
' ***


' command line options must be in the following format:

'   1 "New Project" "Jpeg pictures (*.jpg)|*.jpg|Project (*.mpr)|*.mpr|All files (*.*)|*.*" "c:\users" 1

'   mode "title" "string for filters description and filters" "start directory" "filename" Show_hidden_flag

' _startdir$ is mandatory to avoid read-only dirs errors

Sub __UI_BeforeInit
    TempDirFile$ = _StartDir$ + slash$ + TempFile$
    FileName$ = ""

    ' Quick and easy check of command line parameters
    If _CommandCount < 3 And (Command$(1) < "1" Or Command$(1) > "6") Then
        zz = MessageBox("Invalid or missing argument", "Error", MsgBox_OkOnly)
        FileName = ""
        ExitDialogBox
    End If

    ' if start directory or filename or show_hidden flag is parsed
    StartDirectory$ = _StartDir$
    For i = 4 To _CommandCount 'For i = 4 To 6
        If Len(Command$(i)) > 1 Or Command$(i) = slash$ Then ' Is start directory == "\" or == "/" ?
            If InStr(Command$(i), slash$) Then ' start directory
                StartDirectory$ = Command$(i)
            Else
                FileName = Command$(i) ' filename    ***
            End If
        Else
            If Val(Command$(i)) = 0 Then
                ShowHidden%% = False
            Else
                If Val(Command$(i)) = 1 Then ShowHidden%% = True
            End If
        End If
    Next

    ' check of valid start directory
    On Error GoTo PathError
    ChDir StartDirectory$
    On Error GoTo 0

    If ErrorCode = 76 Then
        FileName = ""
        ExitDialogBox
    End If

End Sub


' ***
' Startup code to set window title
' ***

Sub __UI_OnLoad
    Dim title$(5)

    DialogMode = Val(Command$(1))
    If DialogMode < 0 Or DialogMode > 6 Then DialogModeError

    Do: Loop Until _ScreenExists ' wait for the window to be drawn for first time
    _Title "OpenSaveDialog2"

    If DialogMode > 0 And DialogMode < 5 Then
        title$(1) = "New"
        title$(2) = "Open"
        title$(3) = "Save"
        title$(4) = "Save as"
        Caption(OpenSaveForm) = title$(DialogMode)
    Else
        If Command$(2) = "" Then Caption(OpenSaveForm) = "Select file" Else Caption(OpenSaveForm) = Command$(2)
    End If


    ' show dirs and files
    OpenSaveDialogBox Command$(3)

End Sub


Sub __UI_BeforeUpdateDisplay
    'This event occurs at approximately 30 frames per second.
    'You can change the update frequency by calling SetFrameRate DesiredRate%

End Sub


Sub __UI_BeforeUnload
    'If you set __UI_UnloadSignal = False here you can
    'cancel the user's request to close.

    __UI_UnloadSignal = False
    __UI_Click (CancelButton)

End Sub


Sub __UI_Click (id As Long)

    ' double click simulation
    If Timer - ClickTimer < .3 And id = ClickedControl Then
        DoubleClick%% = True
    Else
        DoubleClick%% = False
    End If

    ClickedControl = id ' preparation for double click
    ClickTimer = Timer

    Select Case id

        Case Form1

        Case AcceptButton
            FileName$ = path$ + slash + Text(FileNameTextBox)

            Select Case DialogMode
                Case 2 ' open
                    If Text(FileNameTextBox) = "" Then
                        i = MessageBox("Invalid filename", "Warning", MsgBox_OkOnly + MsgBox_Exclamation)
                    Else
                        'comprobar si el nombre tecleado o seleccionado existe y mostrar una advertencia
                        If Not _FileExists(FileName$) Then
                            i = MessageBox("File does not exist.", "Warning", MsgBox_OkOnly + MsgBox_Exclamation)
                        Else
                            ExitDialogBox
                        End If
                    End If

                Case Else 'new, save, save as
                    If Text(FileNameTextBox) = "" Then
                        i = MessageBox("Invalid filename", "Warning", MsgBox_Ok + MsgBox_Exclamation)
                    Else
                        'comprobar si el nombre tecleado o seleccionado existe y mostrar una advertencia
                        If _FileExists(FileName$) Then
                            if MessageBox("File already exists.\n\n"+_
                                    "Do you want to overwrite?", "Warning", MsgBox_yesno + MsgBox_Exclamation)=MsgBox_yes _
                            then
                                ExitDialogBox
                            Else
                                FileName$ = ""
                            End If
                        Else
                            ExitDialogBox
                        End If
                    End If
            End Select

        Case FilterList

        Case CancelButton
            FileName$ = ""
            ExitDialogBox

        Case FolderListBox
            'double click in folder list
            If DoubleClick%% And Control(FolderListBox).HoveringVScrollbarButton = 0 And Control(FolderListBox).Value <> 0 Then
                ChangeFolder
            End If

        Case FileListBox
            'double click in file list
            If DoubleClick%% And Control(FileListBox).HoveringVScrollbarButton = 0 And Control(FileListBox).Value <> 0 Then
                Text(FileNameTextBox) = GetItem$(FileListBox, Control(FileListBox).Value)
                __UI_Click (AcceptButton)
            Else If Control(fileListBox1).HoveringVScrollbarButton = 0 And Control(FileListBox).Value <> 0 Then
                    Text(FileNameTextBox) = GetItem$(FileListBox, Control(FileListBox).Value)
                End If
            End If

    End Select
End Sub


Sub __UI_MouseEnter (id As Long)
    Select Case id
        Case Form1

        Case FilterList

        Case AcceptButton

        Case CancelButton

        Case FolderListBox

        Case FileListBox

    End Select
End Sub

Sub __UI_MouseLeave (id As Long)
    '    Select Case id
    '        Case Form1

    '        Case FilterList

    '        Case AcceptButton

    '        Case CancelButton

    '        Case FolderListBox

    '        Case FileListBox

    '    End Select
End Sub

Sub __UI_FocusIn (id As Long)
    Select Case id
        Case PathTB
            TempPath$ = Text(PathTB)

            '        Case FilterList

            '        Case AcceptButton

            '        Case CancelButton

            '        Case FolderListBox

            '        Case FileListBox

    End Select
End Sub

Sub __UI_FocusOut (id As Long)
    'This event occurs right before a control loses focus.
    'To prevent a control from losing focus, set __UI_KeepFocus = True below.
    Select Case id
        Case PathTB
            If TempPath$ <> Text(PathTB) Then
                On Error GoTo PathError 'TempPath$ = Text(PathTB) (FocusIn)
                ChDir Text(PathTB)

                ' update drive letter if on Windows
                $If WIN Then
                    D$ = UCase$(Left$(Text(PathTB), 2))

                    If D$ <> GetItem$(DriveList, Control(DriveList).Value) Then
                        i = 1 ' find list entry value of typed drive letter
                        While D$ <> GetItem$(DriveList, i) And i <= DrivesCount
                            i = i + 1
                        Wend

                        If i > DrivesCount Then
                            PathErrorMessage
                        Else
                            Control(DriveList).Value = i ' will trigger ValueChanged sub
                        End If
                    Else
                        GetFoldersAndFiles ' will not trigger it when shown drive is the same
                    End If
                $Else
                    GetFoldersAndFiles
                $End If

            End If

            '        Case FilterList

            '        Case AcceptButton

            '        Case CancelButton

            '        Case FolderListBox

            '        Case FileListBox

    End Select
End Sub


Sub __UI_MouseDown (id As Long)
    '    Select Case id
    '        Case Form1

    '        Case FilterList

    '        Case AcceptButton

    '        Case CancelButton

    '        Case FolderListBox

    '    End Select
End Sub


Sub __UI_MouseUp (id As Long)
    '    Select Case id
    '        Case Form1

    '        Case FilterList

    '        Case AcceptButton

    '        Case CancelButton

    '        Case FolderListBox

    '        Case FileListBox

    '    End Select
End Sub

Sub __UI_KeyPress (id As Long)
    'When this event is fired, __UI_KeyHit will contain the code of the key hit.
    'You can change it and even cancel it by making it = 0

    Select Case id
        Case PathTB
            If __UI_KeyHit = 13 Then ' if ENTER pressed on path textbox
                SetFocus (FolderListBox) ' change focus and check typed path
                __UI_FocusOut (PathTB) ' this event is not automatically triggered after setfocus command
            End If

            '        Case FilterList

            '        Case AcceptButton

            '        Case CancelButton

        Case FolderListBox
            If __UI_KeyHit = 13 Then ChangeFolder

        Case FileListBox, FileNameTextBox
            If __UI_KeyHit = 13 Then __UI_Click (AcceptButton)

        Case DriveList

    End Select
End Sub

Sub __UI_TextChanged (id As Long)
End Sub

' ***
' Fixes are applied in following routines to prevent __UI_ValuleChanged to be triggered twice (bug)
' ***

Sub __UI_ValueChanged (id As Long)
    Select Case id
        Case DriveList
            $If WIN Then
                If Control(DriveList).Value <> DriveList_Value And Control(DriveList).Value <> 0 Then ' this is to prevent double run this sub (bug)
                    DriveList_Value = Control(DriveList).Value
                    ChangeDrive
                End If
                If Control(DriveList).Value = 0 Then
                    Control(DriveList).Value = DriveList_Value 'StartDriveValue
                End If
            $End If

        Case FilterList
            If Control(FilterList).Value <> FilterList_Value And Control(FilterList).Value <> 0 Then ' this is to prevent double run this sub (bug)
                FilterList_Value = Control(FilterList).Value
                ShowFiles (filter$(Control(FilterList).Value))
            End If

            If Control(FilterList).Value = 0 Then
                Control(FilterList).Value = FilterList_Value
            End If


            '        Case FolderListBox

        Case FileListBox
            If Control(FilterList).Value <> 0 Then
                Text(FileNameTextBox) = GetItem$(FileListBox, Control(FileListBox).Value)
            End If

    End Select
End Sub


Sub __UI_FormResized

End Sub


' ***
' New, Open, Save, Save as funtion
' ***

' INPUT:    Filters$ = String to define filters. MS Windows dialog style

' OUTPUT:   Path+Filename of selected file or empty string if not file selected

Sub OpenSaveDialogBox (Filters$)

    ' fill up drives letter list if working on Windows
    $If WIN Then
        Control(DriveList).Hidden = False ' unhides drive list if on Windows
        StartDrive = Left$(_StartDir$, 2)

        ' drive list init
        i = GetLogicalDriveStringsA(0, DriveString$) 'zero returns the drive string size (bytes)
        DriveString$ = Space$(i) 'set drive string length. Each drive string is NULL terminated (CHR$(0))
        ii = GetLogicalDriveStringsA(i, DriveString$) 'the byte size returns the complete drive string

        For ii = 1 To i Step 4
            drive$ = Mid$(DriveString$, ii, 2)
            If drive$ <> Chr$(0) Then
                AddItem DriveList, drive$
                DrivesCount = DrivesCount + 1
                If drive$ = StartDrive Then StartDriveValue = (ii + 3) / 4 ' drive to show in case of (different) drive error
            End If
        Next
        Control(DriveList).Value = StartDriveValue
        ' end of drive list init
    $End If

    ' fills up filters list box
    f$ = Filters$
    ii = 1
    For i = 1 To Len(Filters$)
        i = InStr(f$, "|")
        AddItem FilterList, Left$(f$, i - 1)

        f$ = Right$(f$, Len(f$) - i)
        i = InStr(f$, "|")
        If i <> 0 Then
            filter$(ii) = Left$(f$, i - 1)
            f$ = Right$(f$, Len(f$) - i)
        Else
            filter$(ii) = f$
            i = Len(Filters$)
        End If
        ii = ii + 1
    Next

    Control(FilterList).Value = 1 ' >>> this triggers SUB ValueChanged

    If DialogMode <> 2 And DialogMode <> 6 Then Text(FileNameTextBox) = FileName ' fill-out filename if needed

    GetFoldersAndFiles ' show folders (in current drive) and filtered files in current folder

End Sub


' ***
' Get dir list and show in listbox
' Get file list and show in filelistbox
' ***

Sub GetFoldersAndFiles

    FoldersCount = 0
    ResetList (FolderListBox) ' clears content up

    fh& = FreeFile
    path$ = _CWD$ ' current working directory
    Text(PathTB) = path$
    If Right$(path$, 1) = slash$ Then
        path$ = Left$(path$, Len(path$) - 1) 'removes slash if on root dir
    Else
        AddItem FolderListBox, ".." ' if not on root dir
    End If

    ' find out if compiling on MS Windows and do console commands to show files and folders
    $If WIN Then
        If Not ShowHidden%% Then Hidden$ = "-h-s"

        'show dirs in dir (folders)
        Shell _Hide "dir /b /ON /AD" + Hidden$ + " >" + Chr$(34) + TempDirFile$ + Chr$(34)

        Open TempDirFile$ For Input As fh&
        Do Until EOF(fh&)
            Line Input #fh&, a$
            If Len(a$) And a$ <> "." Then 'skip blank entries
                'FoldersCount = FoldersCount + 1
                AddItem FolderListBox, a$
            End If
        Loop
        Close fh&

        ' get files in dir
        Shell _Hide "dir  /b /ON /A-D" + Hidden$ + " >" + Chr$(34) + TempDirFile$ + Chr$(34)
    $Else
        ' compiling for Linux and Mac OS
        If ShowHidden%% Then Hidden$ = "-A"

        'get files and dirs and select dirs only
        Shell _Hide "ls -1 " + Hidden$ + " --width=0 -N -p --group-directories-first | grep / >" + Chr$(34) + TempDirFile$ + Chr$(34)

        Open TempDirFile$ For Input As fh&

        While Not EOF(fh&)
        Line Input #fh&, a$
        If Len(a$) And a$ <> "." Then 'skip blank entries and select only dirs (if any)
        AddItem FolderListBox, Left$(a$, Len(a$) - 1) ' do not add "/"
        End If
        Wend
        Close fh&

        'get files and dirs and select files only
        Shell _Hide "ls -1 " + Hidden$ + " --width=0 -N -p --group-directories-first | grep / -v >" + Chr$(34) + TempDirFile$ + Chr$(34)
    $End If

    ' count files in dir
    Open TempDirFile$ For Input As fh&

    Do Until EOF(fh&)
        Line Input #fh&, a$
        If Len(a$) And a$ <> TempFile$ Then 'skip blank entries and Temporary file
            FilesCount = FilesCount + 1
        End If
    Loop

    ' store file list in memory
    ReDim FileList$(FilesCount)

    FilesCount = 1
    Seek fh&, 1

    Do Until EOF(fh&)
        Line Input #fh&, a$
        If Len(a$) And a$ <> TempFile$ Then 'skip blank entries and Temporary file
            FileList$(FilesCount) = a$
            FilesCount = FilesCount + 1
        End If
    Loop

    Close #fh&

    ShowFiles (filter$(Control(FilterList).Value)) ' show filtered files in dir

End Sub


' ***
' show filtered filenames in file list box
' ***

' "Graphic files (*.png;*.bmp;*.png)|*.png;*.bmp;*.png|Project (*.mpr)|*.mpr|All files (*.*)|*.*"

Sub ShowFiles (f$)
    ResetList (FileListBox)

    filter1$ = f$ ' this is to avoid filter$(n) entry to be erased

    If filter1$ = "*.*" Then filter1$ = "" ' Is filter="*.*" ?
    filter1$ = filter1$ + ";"

    ' split several filters in (multi) filter string
    Do
        sc = InStr(filter1$, ";")
        filter2$ = LCase$(Mid$(filter1$, 2, sc - 2))
        filter1$ = Right$(filter1$, Len(filter1$) - sc)

        ' filter and show files in FileList
        For i = 1 To FilesCount - 1
            If LCase$(Right$(FileList$(i), Len(filter2$))) = filter2$ Then
                AddItem FileListBox, FileList$(i)
            End If
        Next
    Loop While InStr(filter1$, ";")

End Sub


' ***
' Save filename or blank and exit from dialog box
' ***

Sub ExitDialogBox
    fh& = FreeFile
    Open TempDirFile$ For Output As fh&
    Print #fh&, FileName$
    Close fh&

    System

End Sub


' ***
' Change to selected folder in FolderList
' ***
Sub ChangeFolder
    ChDir GetItem$(FolderListBox, Control(FolderListBox).Value)
    Call GetFoldersAndFiles ' shows dir and files in dir

End Sub


' ***
' Refresh drive letter and try to change to that drive
' ***

$If WIN Then
    Sub ChangeDrive
        On Error GoTo DriveError ' if drive error, will revert DriveList to last used drive

        ' in case of chdir error, CurrentDrive will be restored to start drive
        ChDir GetItem$(DriveList, DriveList_Value)

        On Error GoTo 0 ' disables error trap
        ErrorCode = 0

        GetFoldersAndFiles
    End Sub
$End If

$If WIN Then
    Sub DriveErrorMessage
        z = MessageBox("Drive error " + GetItem$(DriveList, Control(DriveList).Value), "Error", MsgBox_OkOnly + MsgBox_Exclamation)
        Control(DriveList).Value = StartDriveValue
    End Sub
$End If

Sub PathErrorMessage
    z = MessageBox("Wrong path error\n" + Text(PathTB), "Error", MsgBox_OkOnly + MsgBox_Exclamation)
    Text(PathTB) = TempPath$ ' restores previous path text
End Sub


Sub UnhandledErrorMessage
    z = MessageBox("Unhandled error " + Str$(Err), "Error", MsgBox_OkOnly + MsgBox_Exclamation)
    FileName$ = ""
    ExitDialogBox
End Sub


Sub DialogModeError
    z = MessageBox("Dialog mode error "+_
                    "\nDialog mode="+str$(dialogmode) + Str$(Err), "Error", MsgBox_OkOnly + MsgBox_Exclamation)
    FileName$ = ""
    ExitDialogBox
End Sub



'$INCLUDE:'InForm\InForm.ui'
